@@ -17,6 +17,7 @@
 #include "soup-message.h"
 #include "soup-session.h"
 #include "soup-session-feature.h"
+#include "soup-socket.h"
 #include "soup-uri.h"
 
 /**
@@ -34,8 +35,8 @@
  * and then attach it to a session (or multiple sessions) with
  * soup_session_add_feature().
  *
- * By default, the debugging output is sent to %stdout, and looks
- * something like:
+ * By default, the debugging output is sent to
+ * <literal>stdout</literal>, and looks something like:
  *
  * <informalexample><screen>
  * > POST /unauth HTTP/1.1
@@ -309,7 +310,7 @@ soup_logger_set_response_filter (SoupLogger       *logger,
  * @destroy: a #GDestroyNotify to free @printer_data
  *
  * Sets up an alternate log printing routine, if you don't want
- * the log to go to %stdout.
+ * the log to go to <literal>stdout</literal>.
  **/
 void
 soup_logger_set_printer (SoupLogger        *logger,
@@ -621,6 +622,8 @@ static void
 request_queued (SoupSessionFeature *logger, SoupSession *session,
 		SoupMessage *msg)
 {
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+
 	g_signal_connect (msg, "got-informational",
 			  G_CALLBACK (got_informational),
 			  logger);
@@ -637,6 +640,10 @@ request_started (SoupSessionFeature *feature, SoupSession *session,
 	gboolean restarted;
 	guint msg_id;
 
+	g_return_if_fail (SOUP_IS_SESSION (session));
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	g_return_if_fail (SOUP_IS_SOCKET (socket));
+
 	msg_id = soup_logger_get_id (logger, msg);
 	if (msg_id)
 		restarted = TRUE;
@@ -659,6 +666,8 @@ static void
 request_unqueued (SoupSessionFeature *logger, SoupSession *session,
 		  SoupMessage *msg)
 {
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+
 	g_signal_handlers_disconnect_by_func (msg, got_informational, logger);
 	g_signal_handlers_disconnect_by_func (msg, got_body, logger);
 }
