@@ -79,6 +79,7 @@ static guint signals[LAST_SIGNAL] = { 0 };
 struct SoupClientContext {
 	SoupServer     *server;
 	SoupSocket     *sock;
+	SoupMessage    *msg;
 	SoupAuthDomain *auth_domain;
 	char           *auth_user;
 
@@ -94,7 +95,7 @@ typedef struct {
 } SoupServerHandler;
 
 typedef struct {
-	SoupAddress       *interface;
+	SoupAddress       *iface;
 	guint              port;
 
 	char              *ssl_cert_file, *ssl_key_file;
@@ -105,7 +106,7 @@ typedef struct {
 	GMainLoop         *loop;
 
 	SoupSocket        *listen_sock;
-	GSList            *client_socks;
+	GSList            *clients;
 
 	gboolean           raw_paths;
 	SoupPathMap       *handlers;
@@ -163,8 +164,8 @@ finalize (GObject *object)
 	SoupServerPrivate *priv = SOUP_SERVER_GET_PRIVATE (server);
 	GSList *iter;
 
-	if (priv->interface)
-		g_object_unref (priv->interface);
+	if (priv->iface)
+		g_object_unref (priv->iface);
 
 	g_free (priv->ssl_cert_file);
 	g_free (priv->ssl_key_file);
@@ -176,12 +177,19 @@ finalize (GObject *object)
 	if (priv->listen_sock)
 		g_object_unref (priv->listen_sock);
 
-	while (priv->client_socks) {
-		SoupSocket *sock = priv->client_socks->data;
+	while (priv->clients) {
+		SoupClientContext *client = priv->clients->data;
+		SoupSocket *sock = g_object_ref (client->sock);
+
+		priv->clients = g_slist_remove (priv->clients, client);
+
+		if (client->msg) {
+			soup_message_set_status (client->msg, SOUP_STATUS_IO_ERROR);
+			soup_message_io_finished (client->msg);
+		}
 
 		soup_socket_disconnect (sock);
-		priv->client_socks =
-			g_slist_remove (priv->client_socks, sock);
+		g_object_unref (sock);
 	}
 
 	if (priv->default_handler)
@@ -421,7 +429,7 @@ soup_server_class_init (SoupServerClass *server_class)
 	 * holes.
 	 *
 	 * As with #SoupSession:user_agent, if you set a
-	 * %server_header property that has trailing whitespace,
+	 * #SoupServer:server_header property that has trailing whitespace,
 	 * #SoupServer will append its own product token (eg,
 	 * "<literal>libsoup/2.3.2</literal>") to the end of the
 	 * header for you.
@@ -454,8 +462,8 @@ constructor (GType                  type,
 		return NULL;
 	priv = SOUP_SERVER_GET_PRIVATE (server);
 
-	if (!priv->interface) {
-		priv->interface =
+	if (!priv->iface) {
+		priv->iface =
 			soup_address_new_any (SOUP_ADDRESS_FAMILY_IPV4,
 					      priv->port);
 	}
@@ -471,7 +479,7 @@ constructor (GType                  type,
 	}
 
 	priv->listen_sock =
-		soup_socket_new (SOUP_SOCKET_LOCAL_ADDRESS, priv->interface,
+		soup_socket_new (SOUP_SOCKET_LOCAL_ADDRESS, priv->iface,
 				 SOUP_SOCKET_SSL_CREDENTIALS, priv->ssl_creds,
 				 SOUP_SOCKET_ASYNC_CONTEXT, priv->async_context,
 				 NULL);
@@ -483,10 +491,10 @@ constructor (GType                  type,
 	/* Re-resolve the interface address, in particular in case
 	 * the passed-in address had SOUP_ADDRESS_ANY_PORT.
 	 */
-	g_object_unref (priv->interface);
-	priv->interface = soup_socket_get_local_address (priv->listen_sock);
-	g_object_ref (priv->interface);
-	priv->port = soup_address_get_port (priv->interface);
+	g_object_unref (priv->iface);
+	priv->iface = soup_socket_get_local_address (priv->listen_sock);
+	g_object_ref (priv->iface);
+	priv->port = soup_address_get_port (priv->iface);
 
 	return server;
 }
@@ -503,11 +511,11 @@ set_property (GObject *object, guint prop_id,
 		priv->port = g_value_get_uint (value);
 		break;
 	case PROP_INTERFACE:
-		if (priv->interface)
-			g_object_unref (priv->interface);
-		priv->interface = g_value_get_object (value);
-		if (priv->interface)
-			g_object_ref (priv->interface);
+		if (priv->iface)
+			g_object_unref (priv->iface);
+		priv->iface = g_value_get_object (value);
+		if (priv->iface)
+			g_object_ref (priv->iface);
 		break;
 	case PROP_SSL_CERT_FILE:
 		priv->ssl_cert_file =
@@ -557,7 +565,7 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_uint (value, priv->port);
 		break;
 	case PROP_INTERFACE:
-		g_value_set_object (value, priv->interface);
+		g_value_set_object (value, priv->iface);
 		break;
 	case PROP_SSL_CERT_FILE:
 		g_value_set_string (value, priv->ssl_cert_file);
@@ -690,6 +698,7 @@ soup_client_context_cleanup (SoupClientContext *client)
 		g_free (client->auth_user);
 		client->auth_user = NULL;
 	}
+	client->msg = NULL;
 }
 
 static SoupClientContext *
@@ -770,6 +779,15 @@ got_headers (SoupMessage *req, SoupClientContext *client)
 
 		uri = soup_message_get_uri (req);
 		decoded_path = soup_uri_decode (uri->path);
+
+		if (strstr (decoded_path, "/../") ||
+		    g_str_has_suffix (decoded_path, "/..")) {
+			/* Introducing new ".." segments is not allowed */
+			g_free (decoded_path);
+			soup_message_set_status (req, SOUP_STATUS_BAD_REQUEST);
+			return;
+		}
+
 		soup_uri_set_path (uri, decoded_path);
 		g_free (decoded_path);
 	}
@@ -863,6 +881,8 @@ start_request (SoupServer *server, SoupClientContext *client)
 	msg = g_object_new (SOUP_TYPE_MESSAGE,
 			    SOUP_MESSAGE_SERVER_SIDE, TRUE,
 			    NULL);
+	client->msg = msg;
+
 	if (priv->server_header) {
 		soup_message_headers_append (msg->response_headers, "Server",
 					     priv->server_header);
@@ -880,12 +900,12 @@ start_request (SoupServer *server, SoupClientContext *client)
 }
 
 static void
-socket_disconnected (SoupSocket *sock, SoupServer *server)
+socket_disconnected (SoupSocket *sock, SoupClientContext *client)
 {
-	SoupServerPrivate *priv = SOUP_SERVER_GET_PRIVATE (server);
+	SoupServerPrivate *priv = SOUP_SERVER_GET_PRIVATE (client->server);
 
-	priv->client_socks = g_slist_remove (priv->client_socks, sock);
-	g_signal_handlers_disconnect_by_func (sock, socket_disconnected, server);
+	priv->clients = g_slist_remove (priv->clients, client);
+	g_signal_handlers_disconnect_by_func (sock, socket_disconnected, client);
 	g_object_unref (sock);
 }
 
@@ -897,9 +917,9 @@ new_connection (SoupSocket *listner, SoupSocket *sock, gpointer user_data)
 	SoupClientContext *client;
 
 	client = soup_client_context_new (server, g_object_ref (sock));
-	priv->client_socks = g_slist_prepend (priv->client_socks, sock);
+	priv->clients = g_slist_prepend (priv->clients, client);
 	g_signal_connect (sock, "disconnected",
-			  G_CALLBACK (socket_disconnected), server);
+			  G_CALLBACK (socket_disconnected), client);
 	start_request (server, client);
 }
 
@@ -990,6 +1010,38 @@ soup_server_quit (SoupServer *server)
 		g_main_loop_quit (priv->loop);
 }
 
+/**
+ * soup_server_disconnect:
+ * @server: a #SoupServer
+ *
+ * Stops processing for @server and closes its socket. This implies
+ * the effects of soup_server_quit(), but additionally closes the
+ * listening socket.  Note that messages currently in progress will
+ * continue to be handled, if the main loop associated with the
+ * server is resumed or kept running.
+ *
+ * After calling this function, @server is no longer functional, so it
+ * has nearly the same effect as destroying @server entirely. The
+ * function is thus useful mainly for language bindings without
+ * explicit control over object lifetime.
+ **/
+void
+soup_server_disconnect (SoupServer *server)
+{
+	SoupServerPrivate *priv;
+
+	g_return_if_fail (SOUP_IS_SERVER (server));
+	priv = SOUP_SERVER_GET_PRIVATE (server);
+
+	soup_server_quit (server);
+
+	if (priv->listen_sock) {
+		soup_socket_disconnect (priv->listen_sock);
+		g_object_unref (priv->listen_sock);
+		priv->listen_sock = NULL;
+	}
+}
+
 /**
  * soup_server_get_async_context:
  * @server: a #SoupServer
@@ -1154,8 +1206,9 @@ soup_client_context_get_auth_user (SoupClientContext *client)
  * @user_data: the data passed to @soup_server_add_handler
  *
  * A callback used to handle requests to a #SoupServer. The callback
- * will be invoked after receiving the request body; @msg's %method,
- * %request_headers, and %request_body fields will be filled in.
+ * will be invoked after receiving the request body; @msg's
+ * #SoupMessage:method, #SoupMessage:request_headers, and
+ * #SoupMessage:request_body fields will be filled in.
  *
  * @path and @query contain the likewise-named components of the
  * Request-URI, subject to certain assumptions. By default,
@@ -1191,7 +1244,7 @@ soup_client_context_get_auth_user (SoupClientContext *client)
  *
  * To send the response body a bit at a time using "chunked" encoding,
  * first call soup_message_headers_set_encoding() to set
- * %SOUP_ENCODING_CHUNKED on the %response_headers. Then call
+ * %SOUP_ENCODING_CHUNKED on the #SoupMessage:response_headers. Then call
  * soup_message_body_append() (or soup_message_body_append_buffer())
  * to append each chunk as it becomes ready, and
  * soup_server_unpause_message() to make sure it's running. (The
