@@ -4,8 +4,7 @@
 #include <string.h>
 #include <unistd.h>
 
-#include "libsoup/soup-message.h"
-#include "libsoup/soup-headers.h"
+#include <libsoup/soup.h>
 
 #include "test-utils.h"
 
@@ -243,6 +242,17 @@ static struct RequestTest {
 	  }
 	},
 
+	/* qv bug 579318, do_bad_header_tests() below */
+	{ "Req w/ mangled header",
+	  "GET / HTTP/1.1\r\nHost: example.com\r\nFoo one\r\nBar: two\r\n", -1,
+	  SOUP_STATUS_OK,
+	  "GET", "/", SOUP_HTTP_1_1,
+	  { { "Host", "example.com" },
+	    { "Bar", "two" },
+	    { NULL }
+	  }
+	},
+
 	/************************/
 	/*** INVALID REQUESTS ***/
 	/************************/
@@ -303,13 +313,6 @@ static struct RequestTest {
 	  { { NULL } }
 	},
 
-	{ "Header line with no ':'",
-	  "GET / HTTP/1.1\r\nHost example.com\r\n", -1,
-	  SOUP_STATUS_BAD_REQUEST,
-	  NULL, NULL, -1,
-	  { { NULL } }
-	},
-
 	{ "No terminating CRLF",
 	  "GET / HTTP/1.1\r\nHost: example.com", -1,
 	  SOUP_STATUS_BAD_REQUEST,
@@ -317,27 +320,6 @@ static struct RequestTest {
 	  { { NULL } }
 	},
 
-	{ "Blank line before headers",
-	  "GET / HTTP/1.1\r\n\r\nHost: example.com\r\n", -1,
-	  SOUP_STATUS_BAD_REQUEST,
-	  NULL, NULL, -1,
-	  { { NULL } }
-	},
-
-	{ "Blank line in headers",
-	  "GET / HTTP/1.1\r\nHost: example.com\r\n\r\nConnection: close\r\n", -1,
-	  SOUP_STATUS_BAD_REQUEST,
-	  NULL, NULL, -1,
-	  { { NULL } }
-	},
-
-	{ "Blank line after headers",
-	  "GET / HTTP/1.1\r\nHost: example.com\r\nConnection: close\r\n\r\n", -1,
-	  SOUP_STATUS_BAD_REQUEST,
-	  NULL, NULL, -1,
-	  { { NULL } }
-	},
-
 	{ "Unrecognized expectation",
 	  "GET / HTTP/1.1\r\nHost: example.com\r\nExpect: the-impossible\r\n", -1,
 	  SOUP_STATUS_EXPECTATION_FAILED,
@@ -468,6 +450,7 @@ static struct ResponseTest {
 	  }
 	},
 
+	/* Shoutcast support */
 	{ "Shoutcast server not-quite-HTTP",
 	  "ICY 200 OK\r\nFoo: bar\r\n", -1,
 	  SOUP_HTTP_1_0, SOUP_STATUS_OK, "OK",
@@ -476,6 +459,24 @@ static struct ResponseTest {
 	  }
 	},
 
+	/* qv bug 579318, do_bad_header_tests() below */
+	{ "Response w/ mangled header",
+	  "HTTP/1.1 200 ok\r\nFoo: one\r\nBar two:2\r\nBaz: three\r\n", -1,
+	  SOUP_HTTP_1_1, SOUP_STATUS_OK, "ok",
+	  { { "Foo", "one" },
+	    { "Baz", "three" },
+	    { NULL }
+	  }
+	},
+
+	/* qv bug 602863 */
+	{ "HTTP 1.1 response with leading line break",
+	  "\nHTTP/1.1 200 ok\r\nFoo: bar\r\n", -1,
+	  SOUP_HTTP_1_1, SOUP_STATUS_OK, "ok",
+	  { { "Foo", "bar" },
+	    { NULL } }
+	},
+
 	/*************************/
 	/*** INVALID RESPONSES ***/
 	/*************************/
@@ -820,41 +821,51 @@ do_qvalue_tests (void)
 	}
 }
 
-#define RFC2231_TEST_FILENAME "t\xC3\xA9st.txt"
-#define RFC2231_TEST_HEADER "attachment; filename*=UTF-8''t%C3%A9st.txt"
+#define RFC5987_TEST_FILENAME "t\xC3\xA9st.txt"
+#define RFC5987_TEST_FALLBACK_FILENAME "test.txt"
+
+#define RFC5987_TEST_HEADER_ENCODED  "attachment; filename*=UTF-8''t%C3%A9st.txt"
+
+#define RFC5987_TEST_HEADER_UTF8     "attachment; filename*=UTF-8''t%C3%A9st.txt; filename=\"test.txt\""
+#define RFC5987_TEST_HEADER_ISO      "attachment; filename=\"test.txt\"; filename*=iso-8859-1''t%E9st.txt"
+#define RFC5987_TEST_HEADER_FALLBACK "attachment; filename*=Unknown''t%FF%FF%FFst.txt; filename=\"test.txt\""
 
 static void
-do_rfc2231_tests (void)
+do_content_disposition_tests (void)
 {
 	SoupMessageHeaders *hdrs;
 	GHashTable *params;
 	const char *header, *filename;
 	char *disposition;
+	SoupBuffer *buffer;
+	SoupMultipart *multipart;
+	SoupMessageBody *body;
 
-	debug_printf (1, "rfc2231 tests\n");
+	debug_printf (1, "Content-Disposition tests\n");
 
 	hdrs = soup_message_headers_new (SOUP_MESSAGE_HEADERS_MULTIPART);
 	params = g_hash_table_new (g_str_hash, g_str_equal);
-	g_hash_table_insert (params, "filename", RFC2231_TEST_FILENAME);
+	g_hash_table_insert (params, "filename", RFC5987_TEST_FILENAME);
 	soup_message_headers_set_content_disposition (hdrs, "attachment", params);
 	g_hash_table_destroy (params);
 
 	header = soup_message_headers_get_one (hdrs, "Content-Disposition");
-	if (!strcmp (header, RFC2231_TEST_HEADER))
+	if (!strcmp (header, RFC5987_TEST_HEADER_ENCODED))
 		debug_printf (1, "  encoded OK\n");
 	else {
 		debug_printf (1, "  encoding FAILED!\n    expected: %s\n    got:      %s\n",
-			      RFC2231_TEST_HEADER, header);
+			      RFC5987_TEST_HEADER_ENCODED, header);
 		errors++;
 	}
 
+	/* UTF-8 decoding */
 	soup_message_headers_clear (hdrs);
 	soup_message_headers_append (hdrs, "Content-Disposition",
-				     RFC2231_TEST_HEADER);
+				     RFC5987_TEST_HEADER_UTF8);
 	if (!soup_message_headers_get_content_disposition (hdrs,
 							   &disposition,
 							   &params)) {
-		debug_printf (1, "  decoding FAILED!\n    could not parse\n");
+		debug_printf (1, "  UTF-8 decoding FAILED!\n    could not parse\n");
 		errors++;
 		return;
 	}
@@ -862,25 +873,99 @@ do_rfc2231_tests (void)
 
 	filename = g_hash_table_lookup (params, "filename");
 	if (!filename) {
-		debug_printf (1, "  decoding FAILED!\n    could not file filename\n");
+		debug_printf (1, "  UTF-8 decoding FAILED!\n    could not find filename\n");
 		errors++;
-	} else if (strcmp (filename, RFC2231_TEST_FILENAME) != 0) {
-		debug_printf (1, "  decoding FAILED!\n    expected: %s\n    got:      %s\n",
-			      RFC2231_TEST_FILENAME, filename);
+	} else if (strcmp (filename, RFC5987_TEST_FILENAME) != 0) {
+		debug_printf (1, "  UTF-8 decoding FAILED!\n    expected: %s\n    got:      %s\n",
+			      RFC5987_TEST_FILENAME, filename);
 		errors++;
 	} else
-		debug_printf (1, "  decoded OK\n");
-
+		debug_printf (1, "  UTF-8 decoded OK\n");
 	g_hash_table_destroy (params);
+
+	/* ISO-8859-1 decoding */
+	soup_message_headers_clear (hdrs);
+	soup_message_headers_append (hdrs, "Content-Disposition",
+				     RFC5987_TEST_HEADER_ISO);
+	if (!soup_message_headers_get_content_disposition (hdrs,
+							   &disposition,
+							   &params)) {
+		debug_printf (1, "  iso-8859-1 decoding FAILED!\n    could not parse\n");
+		errors++;
+		return;
+	}
+	g_free (disposition);
+
+	filename = g_hash_table_lookup (params, "filename");
+	if (!filename) {
+		debug_printf (1, "  iso-8859-1 decoding FAILED!\n    could not find filename\n");
+		errors++;
+	} else if (strcmp (filename, RFC5987_TEST_FILENAME) != 0) {
+		debug_printf (1, "  iso-8859-1 decoding FAILED!\n    expected: %s\n    got:      %s\n",
+			      RFC5987_TEST_FILENAME, filename);
+		errors++;
+	} else
+		debug_printf (1, "  iso-8859-1 decoded OK\n");
+	g_hash_table_destroy (params);
+
+	/* Fallback */
+	soup_message_headers_clear (hdrs);
+	soup_message_headers_append (hdrs, "Content-Disposition",
+				     RFC5987_TEST_HEADER_FALLBACK);
+	if (!soup_message_headers_get_content_disposition (hdrs,
+							   &disposition,
+							   &params)) {
+		debug_printf (1, "  fallback decoding FAILED!\n    could not parse\n");
+		errors++;
+		return;
+	}
+	g_free (disposition);
+
+	filename = g_hash_table_lookup (params, "filename");
+	if (!filename) {
+		debug_printf (1, "  fallback decoding FAILED!\n    could not find filename\n");
+		errors++;
+	} else if (strcmp (filename, RFC5987_TEST_FALLBACK_FILENAME) != 0) {
+		debug_printf (1, "  fallback decoding FAILED!\n    expected: %s\n    got:      %s\n",
+			      RFC5987_TEST_FALLBACK_FILENAME, filename);
+		errors++;
+	} else
+		debug_printf (1, "  fallback decoded OK\n");
+	g_hash_table_destroy (params);
+
 	soup_message_headers_free (hdrs);
 
+	/* Ensure that soup-multipart always quotes filename (bug 641280) */
+	multipart = soup_multipart_new (SOUP_FORM_MIME_TYPE_MULTIPART);
+	buffer = soup_buffer_new (SOUP_MEMORY_STATIC, "foo", 3);
+	soup_multipart_append_form_file (multipart, "test", "token",
+					 "text/plain", buffer);
+	soup_buffer_free (buffer);
+
+	hdrs = soup_message_headers_new (SOUP_MESSAGE_HEADERS_MULTIPART);
+	body = soup_message_body_new ();
+	soup_multipart_to_message (multipart, hdrs, body);
+	soup_message_headers_free (hdrs);
+	soup_multipart_free (multipart);
+
+	buffer = soup_message_body_flatten (body);
+	soup_message_body_free (body);
+
+	if (strstr (buffer->data, "filename=\"token\""))
+		debug_printf (1, "  SoupMultipart encoded filename correctly\n");
+	else {
+		debug_printf (1, "  SoupMultipart encoded filename incorrectly!\n");
+		errors++;
+	}
+	soup_buffer_free (buffer);
+
 	debug_printf (1, "\n");
 }
 
 #define CONTENT_TYPE_TEST_MIME_TYPE "text/plain"
 #define CONTENT_TYPE_TEST_ATTRIBUTE "charset"
 #define CONTENT_TYPE_TEST_VALUE     "US-ASCII"
-#define CONTENT_TYPE_TEST_HEADER    "text/plain; charset=\"US-ASCII\""
+#define CONTENT_TYPE_TEST_HEADER    "text/plain; charset=US-ASCII"
 
 #define CONTENT_TYPE_BAD_HEADER     "plain text, not text/html"
 
@@ -961,7 +1046,7 @@ struct {
 	{ "five", "test with \xC3\xA1\xC3\xA7\xC4\x89\xC3\xA8\xC3\xB1\xC5\xA3\xC5\xA1" }
 };
 
-#define TEST_PARAMS_RESULT "one=\"foo\", two=\"test with spaces\", three=\"test with \\\"quotes\\\" and \\\\s\", four, five*=UTF-8''test%20with%20%C3%A1%C3%A7%C4%89%C3%A8%C3%B1%C5%A3%C5%A1"
+#define TEST_PARAMS_RESULT "one=foo, two=\"test with spaces\", three=\"test with \\\"quotes\\\" and \\\\s\", four, five*=UTF-8''test%20with%20%C3%A1%C3%A7%C4%89%C3%A8%C3%B1%C5%A3%C5%A1"
 
 static void
 do_append_param_tests (void)
@@ -986,6 +1071,45 @@ do_append_param_tests (void)
 	} else
 		debug_printf (1, "  OK\n");
 	g_string_free (params, TRUE);
+
+	debug_printf (1, "\n");
+}
+
+static const struct {
+	const char *description, *name, *value;
+} bad_headers[] = {
+	{ "Empty name", "", "value" },
+	{ "Name with spaces", "na me", "value" },
+	{ "Name with colon", "na:me", "value" },
+	{ "Name with CR", "na\rme", "value" },
+	{ "Name with LF", "na\nme", "value" },
+	{ "Name with tab", "na\tme", "value" },
+	{ "Value with CR", "name", "val\rue" },
+	{ "Value with LF", "name", "val\nue" },
+	{ "Value with LWS", "name", "val\r\n ue" }
+};
+
+static void
+do_bad_header_tests (void)
+{
+	SoupMessageHeaders *hdrs;
+	int i;
+
+	debug_printf (1, "bad header rejection tests\n");
+
+	hdrs = soup_message_headers_new (SOUP_MESSAGE_HEADERS_MULTIPART);
+	for (i = 0; i < G_N_ELEMENTS (bad_headers); i++) {
+		debug_printf (1, "  %s\n", bad_headers[i].description);
+		expect_warning = TRUE;
+		soup_message_headers_append (hdrs, bad_headers[i].name,
+					     bad_headers[i].value);
+		if (expect_warning) {
+			expect_warning = FALSE;
+			debug_printf (1, "    FAILED: soup_message_headers_append() did not reject it\n");
+			errors++;
+		}
+	}
+	soup_message_headers_free (hdrs);
 }
 
 int
@@ -996,9 +1120,10 @@ main (int argc, char **argv)
 	do_request_tests ();
 	do_response_tests ();
 	do_qvalue_tests ();
-	do_rfc2231_tests ();
+	do_content_disposition_tests ();
 	do_content_type_tests ();
 	do_append_param_tests ();
+	do_bad_header_tests ();
 
 	test_cleanup ();
 	return errors != 0;
